# Author: Berit Schreck

# Call with > source("BrualdiP.R")

# You have to give a matrix (see below), which is named "x"
# Matrix input: column by column and the 1st no. is the # of rows, 2nd # of columns
#x <- matrix (c(1,0,0,1,1,1,1,1,0),3,3)

# -- OR --

# Run on command line: Rscript Brualdi.R <matrix-file>

for(solved in 1){ # loop to exit the program
args <- commandArgs(TRUE) # permit arguments from terminal to be passed to program
mat <- args[1]
# print(mat)

# Read the matrix from the file "matrix.csv"
###x <- read.csv2(mat, sep=";", header=FALSE) # ODER sep=","!!!
x <- read.csv2(mat, sep=",", header=FALSE)

# output-file
output <- file("Brualdi_out.txt", open = "wt")

# Ordering the matrix according to column sum (descending order)
x <- x [,order(colSums(x),decreasing=TRUE)]

# 2 runs: first w the original matrix, second w the transposed matrix
run <- 0

noShifts_min_O <- 0
noShifts_max_O <- 0

while(run < 2){

noShifts_min <- 0
noShifts_max <- 0

if(run == 1){
	# Transpose matrix (tx) and order according to column sums
	tx <- t(x)
	tx <- tx [,order(colSums(tx),decreasing=TRUE)]
	x <- tx
}
print(x)
# print("----")

# Ferrers matrix
print(dim(x)[1])
print(dim(x)[2])
f <- matrix(nrow = dim(x)[1], ncol = dim(x)[2])
# Fill the Ferrers matrix with entries
for (i in 1:dim(f)[1]) {for (j in 1:dim(f)[2]) {f[i,j]=0};i=i+1;} # every entry is zero
for (i in 1:dim(f)[1]) {j=0; while(j<=rowSums(x)[i]){f[i,j]=1;j=j+1;}; i=i+1;} # fill ones into the Ferrers matrix

# calculate the minimal amount of shifts
minShift <- 0
for(l in 1:dim(x)[2]){
	sumF <- 0
	sumX <- 0
	for(m in 1:dim(x)[1]){
		sumX <- sumX+x[m,l]
		sumF <- sumF+f[m,l]
	}
	diff <- sumF-sumX
	if(diff > 0){
		minShift <- minShift+diff
	}
}
print("----")
#cat("minShift: ", minShift)
print(paste0("minShift: ", minShift))
print("----")

# Calculate the Brualdi (BR) shift value and the degree
BR <- 0
degree <- 0
for(i in 1:dim(x)[2]){
	for(j in 1:dim(x)[1]){
		if(x[j,i] != f[j,i]){
			BR <- BR+1
		}
		if(x[j,i] == 1){
			degree <- degree+1
		}
	}
}
BR <- BR/2
print(paste0("Brualdi: ", BR))
print("----")
print(paste0("Degree: ", degree))
print("----")

# Determine if there are different column sums in the matrix
solution_min <- matrix(nrow = dim(x)[2], ncol = dim(x)[2])
solution_max <- matrix(nrow = dim(x)[2], ncol = dim(x)[2])
solution <- matrix(nrow = dim(x)[2], ncol = dim(x)[2])
columnSums <- colSums(x)
solved <- TRUE
for (i in 1:(dim(x)[2]-1)){
	if(columnSums[i] == columnSums[i+1]){
		solved <- FALSE
		break
	}
}

if (solved == TRUE) {
	# cat(paste0(degree," ",minShift," ", BR," ", BR,"\n"), file=output) # BR = noShifts
	print("EXIT! All columns have different column sums. Solution already given!")

#	if(run == 0){
#		noShifts_min_O <- BR
#		noShifts_max_O <- BR
#		minShift_O <- minShift
#		BR_O <- BR
#	}
#	else{
#		print(BR)
#		noShifts_min <- BR
#		noShifts_max <- BR
#	}
	#break # exits the program if all columns of the matrix have distinct column sums
}

# Determine if all columns of the Ferrer's matrix are the same
solved <- TRUE
for(a in 1:(dim(f)[2]-1)){
	for(b in (a+1):dim(f)[2]){
		for(c in 1:dim(f)[1]){
			if(f[c,a] != f[c,b]){
				solved <- FALSE
				break
			}
		}
		if(!solved) break
	}
	if(!solved) break
}
if (solved == TRUE) {
	# cat(paste0(degree," ",minShift," ", BR," ", BR,"\n"), file=output) # BR = noShifts
	print("EXIT! All columns of the Ferrer's matrix are the same. Solution already given!")

#	if(run == 0){
#		noShifts_min_O <- BR
#		noShifts_max_O <- BR
#		minShift_O <- minShift
#		BR_O <- BR
#	}
#	else{
#		noShifts_min <- BR
#		noShifts_max <- BR
#	}
	#break # exits the program if all columns of the Ferrer's matrix are the same
}

# -- MINIMUM PERFECT WEIGHTED MATCHING --

# Compare the matrix and the corresponding Ferrers matrix column by column and there for every entry
# Take only the columns with same column sums to calculate the order by the perfect weighted matching. The columns with distinct column sums will be aligned by a descending order.
v <- 1
while (v <= dim(x)[2]){
	if (v < dim(x)[2] && columnSums[v] == columnSums[v+1]){
		start <- v
		repeat{
			v <- v+1
			end <- v
			if (v == dim(x)[2] || columnSums[v] != columnSums[v+1]) break
		}
		# A matrix for the weights in the perfect matching problem ... difference between a column of the original matrix and the corresponding Ferrers matrix
		c <- matrix(nrow=end-start+1,ncol=end-start+1) # create the matrix for the weights
		for (i in start:end){
			for (k in start:end){
				s <- 0
				for (j in 1:dim(x)[1]){
					if(x[j,i] != f[j,k]){
						s <- s+1
					}
				}
				c[i-start+1,k-start+1] <- s
			}
		}
		# fill the matrix c with weights
		library(lpSolve) # attach the library to solve the perfect matching problem
		assign.costs <- c
		sol_max <- lp.assign(assign.costs, direction="max")$solution
		sol_min <- lp.assign(assign.costs)$solution # print the solution: if you have a 1 in it then the [i,]-column of the original matrix should be at the [,j]th position
		if (start > 1){
			for (j in 1:start-1){
				for (a in start:end){
					solution_min[j,a] <- 0
					solution_max[j,a] <- 0
				}
			}
		}
		for (b in start:end){
			for (j in start:end){
				solution_min[j,b] <- sol_min[j-start+1,b-start+1]
				solution_max[j,b] <- sol_max[j-start+1,b-start+1]
			}
		}
		if (end < dim(x)[2]){
			j <- end+1
			while (j <= dim(x)[2]){
				for (c in start:end){
					solution_min[j,c] <- 0
					solution_max[j,c] <- 0
				}
				j <- j+1
			}
		}
	}
	else {
		for (j in 1:dim(x)[2]){
			solution_min[j,v] <- 0
			solution_max[j,v] <- 0
			if (j == v){
				solution_min[j,v] <- 1
				solution_max[j,v] <- 1
			}
		}
	}
	v <- v+1
}
# print(solution_min)
# print(solution_max)

# Get the matrix which needs the smallest amount of shift to get nested
solMatrix_min <- matrix(nrow=dim(x)[1], ncol=dim(x)[2])
solMatrix_max <- matrix(nrow=dim(x)[1], ncol=dim(x)[2])
for (i in 1:dim(x)[2]){
	for (p in 1:dim(x)[2]){
		if(solution_min[p,i] != 0){
			solMatrix_min[,i] <- x[,p]
			break
		}
	}
}
for (i in 1:dim(x)[2]){
	for (p in 1:dim(x)[2]){
		if(solution_max[p,i] != 0){
			solMatrix_max[,i] <- x[,p]
			break
		}
	}
}
# print("solMatrix_min: ")
# print(solMatrix_min)
 print("solMatrix_max: ")
 print(solMatrix_max)

# Calculate the amount of shifts
for (i in 1:dim(x)[2]){
	for(j in 1:dim(x)[1]){
		if(solMatrix_min[j,i] != f[j,i]){
			noShifts_min <- noShifts_min+1
		}
	}
}
noShifts_min <- noShifts_min/2
print(paste0("#Shifts_min: ", noShifts_min))

for (i in 1:dim(x)[2]){
	for(j in 1:dim(x)[1]){
		if(solMatrix_max[j,i] != f[j,i]){
			noShifts_max <- noShifts_max+1
		}
	}
}
noShifts_max <- noShifts_max/2
print(paste0("#Shifts_max: ", noShifts_max))

# Write results to variables for second run (transposed matrix)
if (run == 0){
	minShift_O <- minShift
	BR_O <- BR
	noShifts_min_O <- noShifts_min
	noShifts_max_O <- noShifts_max
}

run <- run+1
}

# Determin whether to use the results from the original or transposed matrix
original = FALSE
best_min = noShifts_min_O
if (noShifts_min < noShifts_min_O) {
	best_min = noShifts_min
	original = TRUE
}
best_max = noShifts_max_O
if (noShifts_max > noShifts_max_O && original) {
	best_max = noShifts_max
}

# Write results to file
cat(paste0(degree," ",minShift_O," ", BR_O," ", noShifts_min_O," ", noShifts_max_O," ",minShift," ", BR," ", noShifts_min," ", noShifts_max," ", best_min," ",best_max,"\n"), file=output)
}
